@@ -20,6 +20,20 @@
 typedef struct _context _context;
 typedef struct _clock _clock;
 
+#ifdef NEKO_NO_THREADS
+
+#define context_new() NULL
+#define context_delete(ctx)
+#define context_set(ctx,v)	ctx = ((_context*)(v))
+#define context_get(ctx)	((void*)(ctx))
+
+#define context_lock_new()	NULL
+#define context_lock(l)
+#define context_release(l)
+#define context_lock_delete(l)
+
+#else
+
 _context *context_new();
 void context_delete( _context *ctx );
 void context_set( _context *ctx, void *v );
@@ -30,5 +44,7 @@ void context_lock( _clock *l );
 void context_release( _clock *l );
 void context_lock_delete( _clock *l );
 
+#endif
+
 #endif
 /* ************************************************************************ */
