@@ -50,6 +50,9 @@ struct _neko_vm {
 	custom_list *clist;
 	value resolver;
 	char tmp[100];
+	int trusted_code;
+	neko_stat_func fstats;
+	neko_stat_func pstats;
 };
 
 extern int_val *callback_return;
@@ -59,6 +62,7 @@ extern _context *neko_vm_context;
 
 extern value neko_alloc_apply( int nargs, value env );
 extern value neko_interp( neko_vm *vm, void *m, int_val acc, int_val *pc );
+extern int_val *neko_get_ttable();
 
 #endif
 /* ************************************************************************ */
